/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.data.input;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.jaxen.JaxenException;
import org.jaxen.XPath;
import org.jaxen.jdom.JDOMXPath;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import monalisa.data.pn.Arc;
import monalisa.data.pn.PetriNet;
import monalisa.data.pn.Place;
import monalisa.data.pn.Transition;
import monalisa.data.pn.UniquePetriNetEntity;
import monalisa.util.FileUtils;

public class SppedInputHandler implements InputHandler {
    private final Map<Integer, Place> places = new HashMap<Integer, Place>();
    private final Map<Integer, Transition> transitions = new HashMap<Integer, Transition>();

    @Override
    public boolean isKnownFile(File file) throws IOException {
        return "spped".equalsIgnoreCase(FileUtils.getExtension(file));
    }

    @Override
    public PetriNet load(InputStream in) throws IOException {
        SAXBuilder builder = new SAXBuilder();
        Document doc;
        
        try {
            doc = builder.build(in);
        } catch (JDOMException e) {
            throw new IOException("Failed to parse the XML file.", e);
        }
        
        Element root = doc.getRootElement();
        
        final String expectedVersion = "2";
        
        if (!"Snoopy".equals(root.getName()) ||
                !expectedVersion.equals(root.getAttributeValue("version")))
            throw new IOException("Spped file has wrong format/version.");
        
        try {
            return parse(doc);
        } catch (JDOMException e) {
            throw new IOException(e);
        } catch (JaxenException e) {
            throw new IOException(e);
        }
    }
    
    @SuppressWarnings("unchecked")
    private PetriNet parse(Document doc) throws JDOMException, JaxenException {
        PetriNet petriNet = new PetriNet();

        Map<Integer, UniquePetriNetEntity> entities =
            new HashMap<Integer, UniquePetriNetEntity>();
        final String nodePath = "/Snoopy/nodeclasses/nodeclass[@name='%s']/node";
        final XPath placesPath = new JDOMXPath(String.format(nodePath, "Place"));
        final XPath transitionPath = new JDOMXPath(String.format(nodePath, "Transition"));

        petriNet.putProperty("name",
            doc.getRootElement().getChild("netclass").getAttributeValue("name"));

        List<Element> placeNodes = placesPath.selectNodes(doc);
        Place place;
        int countPlaces = 0, id, internalId;
        Long marking;
        String name;
        for (Element placeNode : placeNodes) {
            id = Integer.parseInt(getAttributeValue(placeNode, "ID"));
            internalId = Integer.parseInt(placeNode.getAttributeValue("id"));
            name = getAttributeValue(placeNode, "Name");
            if(name == null || name.isEmpty())
                name = Integer.toString(id);
            marking = Long.parseLong(getAttributeValue(placeNode, "Marking"));
            place = findPlace(countPlaces, petriNet);
            place.putProperty("name", name);
            place.putProperty("posX", new Double(((Element)placeNode.getChild("graphics").getChildren().toArray()[0]).getAttributeValue("x")));
            place.putProperty("posY", new Double(((Element)placeNode.getChild("graphics").getChildren().toArray()[0]).getAttributeValue("y")));
            petriNet.addPlace(place);
            petriNet.setTokens(place, marking);
            entities.put(internalId, place);
            countPlaces++;
        }

        List<Element> transitionNodes = transitionPath.selectNodes(doc);
        Transition transition;
        int countTransitions = 0;
        for (Element transitionNode : transitionNodes) {
            id = Integer.parseInt(getAttributeValue(transitionNode, "ID"));
            internalId = Integer.parseInt(transitionNode.getAttributeValue("id"));
            name = getAttributeValue(transitionNode, "Name");
            if(name == null || name.isEmpty())
                name = Integer.toString(id);
            transition = findTransition(countTransitions, petriNet);
            transition.putProperty("name", name);
            transition.putProperty("posX", new Double(((Element)transitionNode.getChild("graphics").getChildren().toArray()[0]).getAttributeValue("x")));
            transition.putProperty("posY", new Double(((Element)transitionNode.getChild("graphics").getChildren().toArray()[0]).getAttributeValue("y")));
            petriNet.addTransition(transition);
            entities.put(internalId, transition);
            countTransitions++;
        }

        final XPath edgePath = new JDOMXPath("/Snoopy/edgeclasses/edgeclass[@name='Edge']/edge");

        List<Element> edgeNodes = edgePath.selectNodes(doc);
        for (Element edgeNode : edgeNodes) {
            int fromId = Integer.parseInt(edgeNode.getAttributeValue("source"));
            int toId = Integer.parseInt(edgeNode.getAttributeValue("target"));
            int weight = Integer.parseInt(getAttributeValue(edgeNode, "Multiplicity"));
            Arc arc = new Arc(entities.get(fromId), entities.get(toId), weight);

            UniquePetriNetEntity from = entities.get(fromId);
            if (from instanceof Place) {
                Arc oldArc = petriNet.getArc((Place) from, (Transition) entities.get(toId));
                if (oldArc != null)
                    oldArc.setWeight(oldArc.weight() + 1);
                else
                    petriNet.addArc((Place) from, (Transition) entities.get(toId), arc);
            }
            else {
                Arc oldArc = petriNet.getArc((Transition) from, (Place) entities.get(toId));
                if (oldArc != null)
                    oldArc.setWeight(oldArc.weight() + 1);
                else
                    petriNet.addArc((Transition) from, (Place) entities.get(toId), arc);
            }
        }

        return petriNet;
    }
    
    private String getAttributeValue(Element node, String attribute) throws JaxenException {
        XPath selector = new JDOMXPath("attribute[@name='" + attribute + "']");
        return ((Element) selector.selectSingleNode(node)).getTextTrim();
    }

    private Place findPlace(int placeId, PetriNet petriNet) {
        Place place = places.get(placeId);

        if (place == null) {
            place = new Place(placeId);
            places.put(placeId, place);
            petriNet.addPlace(place);
        }
        return place;
    }

    private Transition findTransition(int transitionId, PetriNet petriNet) {
        Transition transition = transitions.get(transitionId);

        if (transition == null) {
            transition = new Transition(transitionId);
            transitions.put(transitionId, transition);
            petriNet.addTransition(transition);
        }
        return transition;
    }
}
