/*
 *
 *  This file ist part of the software MonaLisa.
 *  MonaLisa is free software, dependend on non-free software. For more information read LICENCE and README.
 *
 *  (c) Molekulare Bioinformatik, Goethe University Frankfurt, Frankfurt am Main, Germany
 *
 */

package monalisa.tools.cluster;

/*
 * Die Klasse Upgma.java implementiert das hierarchische agglomerative
 * Clusterverfahren UPGMA
 * @author belau
 */
public class UPGMA {
    int cn; // Clusteranzahl innerhalb eines Clusters
    int leftCn; // Clusteranzahl innerhalb des linken Kindes
    int rigthCn; // Clusteranzahl innerhalb des rechten Kindes
    int n; // Anzahl der momentanen Cluster
    float dmin; // Dmin: minimale paarweise Distanz
    int imin; // Dmin: Index i
    int jmin; // Dmin: Index j
    Cluster[] cluster; // array von Cluster-Objekten
    String name; // Clusterlabel
    String member; // T-Inv. innerhalb eines Clusters

    /* Konstruktor der Klasse Upgma */
    public UPGMA(float[][] d, float[][] e, String[] label, float th) {

        int l = d.length - 1; // Anzahl der Blätter

        cluster = new Cluster[2 * l]; // Initialisierung des Cluster-arrays

        /* Initialisierung & Einlesen der Blaetterknoten */
        for (int i = 1; i <= l; i++) {
            name = label[i - 1];
            member = name + ",";
            cluster[i] =
                new Cluster(member, name, i, 1, 0, 0, 0, null, 0, null, 0,
                    1.0f, 0, 0, 0);
        }
        n = l; // momentane Clusteranzahl entspricht Blaetteranzahl

        /* Berechnung von dmin */
        while (n < 2 * l - 1) {
            dmin = 10000f;
            for (int i = 2; i <= l; i++) {
                for (int j = 1; j < i; j++) {
                    float erg = d[i][j];
                    if (erg < dmin) {
                        dmin = erg;
                        imin = i;
                        jmin = j;
                    }
                }
            }
            join(imin, jmin, d, th);
            update(d, e);
        }// while
    }// Konstruktor Upgma

    /*
     * Methode zur Vereinigung der Cluster i und j mit minimaler paarweisen
     * Distanz zu neuem Cluster k
     */
    public void join(int i, int j, float[][] d, float th) {

        ++n; // Vergroeßerung der Clusteranzahl

        if (dmin > th) { // Hierarchieebenen-spezifische Dendrogrammdarstellung:
            // Schwellenwertueberschreitung: Inv. werden nicht in ein
            // Blaettercluster zusammengefaßt

            /* Berechnung der Objektattribute des linken Kindes */
            Cluster left = cluster[(int) d[imin][0]];
            int leftCn = left.getCn();
            String leftName = left.getName();
            float leftHeigth = dmin - (left.getHeigth());

            /* Berechnung der Objektattribute des rechten Kindes */
            Cluster rigth = cluster[(int) d[0][jmin]];
            int rigthCn = rigth.getCn();
            String rightName = rigth.getName();
            float rightHeigth = dmin - (rigth.getHeigth());

            int cn = leftCn + rigthCn;
            String name = leftName + ", " + rightName;
            String member = leftName + ", " + rightName;
            int mitte = 1;
            name = "N" + n;

            /* Deklaration und Initialisierung des neuen Clusters k */
            cluster[n] =
                new Cluster(member, name, n, cn, leftCn, rigthCn, dmin, left,
                    leftHeigth, rigth, rightHeigth, mitte, 0, 0, 0);

        } else { // Hierarchieebenen-spezifische Dendrogrammdarstellung:
            // keine Schwellenwertueberschreitung: Inv. werden zu einem
            // Blaettercluster vereinigt

            Cluster left = cluster[(int) d[imin][0]];
            Cluster rigth = cluster[(int) d[0][jmin]];
            // String leftName = left.getName();
            String leftMember = left.member;
            // String rightName = rigth.getName();
            String rightMember = rigth.member;

            String member = leftMember + " " + rightMember;
            String name = "Inv_" + n;
            cluster[(int) d[imin][0]] = null;
            cluster[(int) d[0][jmin]] = null;

            // Deklaration und Initialisierung des neuen Clusters k
            cluster[n] =
                new Cluster(member, name, n, 1, 0, 0, 0, null, 0, null, 0, 1,
                    0, 0, 0);
        }
    }

    /* Methode zur Neuberechnung der Distanzmatrix */
    public void update(float[][] d, float[][] e) {

        if (imin < jmin) { // Voraussetzung für update imin < jmin
            // EMPTY!
        } else {
            int i = imin;
            imin = jmin;
            jmin = i;
        }

        for (int i = 1; i <= d.length - 1; i++) {

            /*
             * Berechnung der Distanz zwischen dem neuen Cluster k und allen
             * anderen Clustern
             */
            if ((i != imin) && (i != jmin) && (d[imin][i] != 0)) {
                e[imin][i] = (d[imin][i] + d[jmin][i]) / 2;
            } else {
                e[imin][i] = 1000f;
            }
        }

        /* Kopieren von e in d & Symmetrierung der Distanzmatrix */
        for (int i = 1; i <= d.length - 1; i++) {
            d[i][imin] = e[imin][i];
            d[imin][i] = d[i][imin];
        }

        /* Loeschen der Reihe und Spalte von jmin */
        for (int i = 1; i <= d.length - 1; i++) {
            d[i][jmin] = 10000f;
            d[jmin][i] = d[i][jmin];
        }

        /* Update der Beschriftung */
        d[imin][0] = n;
        d[0][imin] = n;
    }// update

    /* Methode zur Uebergabe des Wurzel-Clusters */
    public Cluster getWurzel() {
        return cluster[n];
    }// getWurzel
}// class Upgma